#pragma once
#include "MyTools.h"
#include <unordered_map>
#include <string>
#include <sstream>
#include <fstream>
//#include <iomanip>

/**
*Class which represents highscore database.
*/
class HighScore {
public:
	/**
	*Enum which represents types of result for user on map according to the previous time of user.
	*/
	enum result {
		bad_time, /*! < Current time is worse than the previous.*/
		best_user_time, /*! < The best time of user.*/
		best_map_time /*! < The best time for map.*/
	};

	/**
	*Gets string with the best time for map and with the name of user which beats this time.

	*@param map_id Identifier of map.
	*@return The best time in format hh:mm:ss for map with the name of user which beats this time.
	*/
	std::string get_map_best_time(const std::string & map_id) {
		std::string & best_user = map_highscore[map_id];
		if ("" == best_user) //noone solved this map yet
			return "";		
		std::stringstream ss;
		MyTools::sec_to_string_stream(user_highscore[best_user][map_id], ss);
		return  ss.str() + " by " + best_user;
	}

	/**
	*Gets string with the best user's time for map.
	*
	*@param user Name of user.
	*@param map_id Identifier of map.
	*@return The best time of user on map.
	*/
	std::string get_user_best_time(const std::string & user, const std::string & map_id) {
		uint64_t sec = user_highscore[user][map_id];
		if (!sec) //sec == 0
			return "";
		std::stringstream ss;
		MyTools::sec_to_string_stream(sec, ss);
		return ss.str();
	}
	
	/**
	*Set time for user on map. Time is update if it is better than previous.
	*
	*@param user Name of user.
	*@param map_id Identifier of map.
	*@param time Time of user on map.
	*@see result TODO
	*@return Enum which represents if the best time was beaten and how.
	*/
	result set_time(const std::string & user, const std::string & map_id, uint64_t time) {
		if (0 == time) //no data for user -> it could happend during initialization of highscore
			return bad_time;
		result ret = bad_time;
		std::string & old_user_map_best = map_highscore[map_id]; //get name of user which has high score on this map
		uint64_t map_high_score = user_highscore[old_user_map_best][map_id]; //best highscore on this map - it is necessary to save it at first (due to situation when user beat his own map highscore)

		//update map score for user if it is better than original
		if (0 == user_highscore[user][map_id] || time < user_highscore[user][map_id]) {
			user_highscore[user][map_id] = time;
			ret = best_user_time;
		}
		//update map best score
		if ("" == old_user_map_best) {
			map_highscore[map_id] = user;
			ret = best_map_time;
		}
		else if (map_high_score > user_highscore[user][map_id]) {
			map_highscore[map_id] = user;
			ret = best_map_time;
		}
		return ret;
	}

	/**
	*Writes high score database to the stream.
	*
	*@param ofs Stream which is highscore writen to.
	*/
	void to_stream(std::ofstream & ofs) {
		std::map<std::string, std::map<std::string, std::string>> mymap;

		for (auto const & user : user_highscore) {
			if ("" == user.first)
				continue;
			ofs << user.first;
			for (auto const &map_hs : user.second) {
				ofs << ';' << map_hs.first << ';' << map_hs.second;
			}
			ofs << std::endl;
		}
	}


private:
	/**
	*Unordered map.
	*Represents high score for map, for each map there is name of user which has the best time.
	*/
	std::unordered_map<std::string, std::string> map_highscore; 
	/**
	*Unordered map.
	*represents high score for user, for each user there is a unordered map with map and best time
	*/
	std::unordered_map<std::string, std::unordered_map<std::string, uint64_t>> user_highscore;
	
};