#pragma once
#include "State.h"
#include <fstream> //ifstream
#include "MyExc.h" 

using element = char; //alias for char representing map element

/**
*Class which provides arithmetics operations over map data using map element constants
*@see Map
*/
class MapUtils {
public:
	/**
	*Checks if element is waklable (groud or target)
	*
	*@param el Element which is checked.
	*@return If element el is walkable.
	*/
	static bool is_walkable_element(element el);

	/**
	*Check if element is movable and with no barrel
	*
	*@param el Element which is checked
	*@param barrel_info Information about the barrel for this element. 0 no barrel, otherwise barrel.
	*@return If it is possible to move barrel on this element.
	*/
	static bool is_barrel_movable_element(element el, int barrel_info);

	/**
	*Makes data complete element from gamegrid element and barrel information.
	*
	*@param el Gamegrid element.
	*@param barrel_info Information about the barrel for this element. 0 no barrel, otherwise barrel.
	*@return Data complete game element.
	*/
	static element make_full_element(element el, int barrel_info);

	/**
	*Makes data complete sokoban element from gamegrid element and sokoban element.
	*
	*@param el Gamegrid element
	*@param sokoban_el Sokoban element.
	*@return Data complete game element.
	*/
	static element make_sokoban_element(element el, element sokoban_el);

	/**
	*Checks if elements of corner triangle allows to move barrel or not (dead position).
	*
	*@ el_1 Element of the first part of corner.
	*@ b_1  Barrel information of the first part of corner.
	*@ el_2 Element of the second part of corner.
	*@ b_2 Barrel information of the second part of corner.
	*@ el_3 Element of the third part of corner.
	*@ b_3 Barrel information of the third part of corner.
	*@return True if corner does not allow to move barrel.
	*/
	static bool is_dead_position_triangle(element el_1, int b_1, element el_2, int b_2, element el_3, int b_3);

	/**
	*Count the number of new barrel on target (+1, 0, -1).
	*
	*@param el_of_new_barrel New element which barrel is now sitting on.
	*@param el_of_old_barrel Original element which barrel was sitting on previously.
	*@return 1, 0 or -1 according to the change.
	*/
	static int number_of_new_barrel_on_target(element el_of_new_barrel, element el_of_old_barrel);
};

/**
*Abstract class representing general map (ancestor for two player and single player map) and its api.
*/
class Map {
public:
	/*! \name Constants which represents the number of game elements. They use easy arithmetics which is used in engine.
	*ground + barrel = barrel <br>
	*ground + sokoban = sokoban <br>
	*ground + sokoban2 = sokoban2 <br>
	*target + barrel = barrel target <br>
	*target + sokoban = sokoban target <br>
	*target + sokoban2 = sokoban2 target
	*/
	//@{
	static const element GROUND = 0;
	static const element TARGET = 1;
	static const element BARREL = 2;
	static const element BARREL_TARGET = 3;
	static const element SOKOBAN = 4;
	static const element SOKOBAN_TARGET = 5;
	static const element WALL = 6;
	static const element GRASS = 7;
	static const element SOKOBAN_2 = 8;
	static const element SOKOBAN_2_TARGET = 9;
	//@}

	static const char NUMBER_OF_FIELDS = 10; /*!< Number of game elements.*/

	static const int MAX_STEPS = 1000; /*!< Maximal number of steps in backtracking.*/
	static const char EOM = 'x';

	/**
	*Checks if map is completed (all barrels are on targets).

	@return Bool value which determines if map is completed.
	*/
	virtual bool complete() = 0;

	/**
	*Restarts level.
	*/
	virtual void restart() = 0;
	/**
	*Virtual destructor.
	*/
	virtual ~Map() {}
	/**
	*Gets with of map (horizontal number of cells).
	*@return Width of map in number of cells.
	*/
	int get_width() const { return width; }
	/**
	*Gets with of map (vertical number of cells).
	*@return Height of map in number of cells.
	*/
	int get_height() const { return height; }
	/**
	*Tries to solve map using backtracking.
	*
	*@return Bool value which determines if map was solved.
	*/
	virtual bool solve() = 0;
	/**
	*If map is solved do next step in solution.
	*
	*@return True if map is sloved and steps is done, false if map is not solved yet.
	*/
	virtual bool play_solution_next() = 0;
	/**
	*Gets type of cell.
	*
	*@param x The coordinate x of cell.
	*@param y The coordinate y of cell.
	*@return Byte number which represents field type according to the constants.
	*/
	virtual element get_field_type(int x, int y) = 0;

	/**
	*Determines if in actual state of game is possible to move with all barrels.
	*
	*@return Bool value which determines if in actual state of game is possible to move with all barrels.
	*/
	virtual bool is_dead_position() = 0;
	virtual void right_move() = 0; /*!<Moves the first sokoban to the right if it is possible. */
	virtual void left_move() = 0; /*!<Moves the first sokoban to the left if it is possible.*/
	virtual void up_move() = 0; /*!<Moves the first sokoban up if it is possible.*/
	virtual void down_move() = 0; /*!<Moves the first sokoban down if it is possible.*/
	virtual void right_move_second() = 0; /*!< Moves the second sokoban to the right if it is possible.*/
	virtual void left_move_second() = 0; /*!<Moves the second sokoban to the left if it is possible.*/
	virtual void up_move_second() = 0; /*!<Moves the second sokoban up if it is possible.*/
	virtual void down_move_second() = 0; /*!<Moves the second sokoban down if it is possible.*/
protected:
	/**
	*Constructor. Creates general map from the input stream.
	*
	*@param map_stream Input stream which the map is made from.
	*/
	Map(std::ifstream & map_stream);

	/**
	*Checks if position (x,y) is in map grid.
	*
	*@param x First position coordinate.
	*@param y Second position coordinate.
	*@return True if position (x,y) is in map grid. False otherwise.
	*/
	bool is_valid_position(int x, int y) {
		return 0 <= x && 0 <= y && width > x && height > y;
	}

	std::vector<std::vector<char>> game_grid; /*!<Array which represents game grid.*/
	int number_of_barrels; /*!<Number of barrels in the map.*/
	int width; /*!<Number of horizontal cells of the map.*/
	int height; /*!<Number of vertical fields of the map.*/
	/**
	*Game state which is used by the constructor for general map including a position of sokoban and position of barrels.
	*It is not used in concrete map. Information from it is copied to another variables
	*/
	TwoPlayerGameState init_state; 
};

/**
*Class which represents map for two players. Some of the general methods which are inherrited does nothing.
*
*/
class TwoPlayerMap : public Map {
public:
	/**
	*Constructor. Creates the two player map from the input file. It uses the constructor from ancestor to reach it.
	*
	*@see Map::Map()
	*@param map_stream Input stream which the map is made from. 
	*/
	TwoPlayerMap(std::ifstream & map_stream);

	/**
	*Checks if map is completed (all barrels are on targets).

	@return Bool value which determines if map is completed.
	*/
	bool complete() override {
		return state.barrels_on_target == number_of_barrels;
	}

	/**
	*Restarts level.
	*/
	void restart() override {
		state = start_state;
	}

	/**
	*This method does nothing. It is implemented only in one player map.
	*@todo
	*@return Just false.
	*/
	bool solve() override { return false; }

	/**
	*This method does nothing. It is implemented only in one player map.
	*
	*@todo
	*@return Just false.
	*/
	bool play_solution_next() override { return false; }

	/**
	*Gets type of cell.
	*
	*@param x The coordinate x of cell.
	*@param y The coordinate y of cell.
	*@return Byte number which represents field type according to the constants.
	*/
	element get_field_type(int x, int y) override {
		if (x == state.sokoban_x && y == state.sokoban_y)
			return MapUtils::make_sokoban_element(game_grid[y][x], SOKOBAN);
		if (x == state.sokoban2_x && y == state.sokoban2_y)
			return MapUtils::make_sokoban_element(game_grid[y][x], SOKOBAN_2);
		return MapUtils::make_full_element(game_grid[y][x], state.barrel_grid[y][x]);
	}

	/**
	*This method does nothing. It is implemented only in the single player map.
	*@todo
	*@return Just false.
	*/
	bool is_dead_position() override { return false; }
	void right_move() override { move_sokoban(1, 0); } /*!<Moves the first sokoban to the right if it is possible. */
	void left_move() override { move_sokoban(-1, 0); } /*!<Moves the first sokoban to the left if it is possible.*/
	void up_move() override { move_sokoban(0, -1); } /*!<Moves the first sokoban up if it is possible.*/
	void down_move() override { move_sokoban(0, 1); } /*!<Moves the first sokoban down if it is possible.*/
	void right_move_second() override { move_sokoban(1, 0, false); }  /*!< Moves the second sokoban to the right if it is possible*/
	void left_move_second() override { move_sokoban(-1, 0, false); } /*!<Moves the second sokoban to the left if it is possible*/
	void up_move_second() override { move_sokoban(0, -1, false); }	 /*!<Moves the second sokoban up if it is possible.*/
	void down_move_second() override { move_sokoban(0, 1, false); }	 /*!<Moves the second sokoban down if it is possible.*/

private:
	TwoPlayerGameState state; /*!<State of game including a position of sokoban and position of barrels.*/
	TwoPlayerGameState start_state;	/*!<Default state of game - for restart*/

	/**
	*General method for sokoban movement. The direction is defined by vector.
	*
	*@param v_x The part x of vector which represents move direction of Sokoban. It may be -1,0 or 1.
	*@param v_y The part x of vector which represents move direction of Sokoban. It may be -1,0 or 1.
	*@param first_sokoban Determines if it is the movement of the first sokoban. 
	*@return 0 if Sokoban was moved, -1 if Sokoban cannot be moved, -2 if input vector cannot represent Sokoban movement
	*/
	int move_sokoban(int v_x, int v_y, bool first_sokoban = true);
};
