#pragma once
#include<vector>
#include<fstream>
#include<map>
#include "State.h" //GameState
#include "Map.h" //Map
#include<string>
#include <stdexcept>
#include "MyExc.h" //corrupted_file_error
/**
*	Class which represents single player map.
*/
class SinglePlayerMap : public Map {
public:
	/**
	*Constructor. Creates the sigle player map from the input file. It uses the constructor from ancestor to reach it.
	*
	*@see Map::Map()
	*@throws corrupted_file_error file is in wrong format
	*@param map_stream Input stream which the map is made from.
	*/
	SinglePlayerMap(std::ifstream & map_stream);

	/**
	*Checks if map is completed (all barrels are on targets).

	@return Bool value which determines if map is completed.
	*/
	bool complete() override {
		return state.barrels_on_target == number_of_barrels;
	}

	/**
	*	Restarts level
	*/
	void restart() override {
		state = start_state;
		play_position = solution.size();
	}

	/**
	*Tries to solve map using backtracking.
	*
	*@return Bool value which determines if map was solved.
	*/
	bool solve() override{
		if (solution.empty()) { //puzzle is not solved yet
			restart(); //load start position
			if (!solve(0)) //try to solve, if false -> return false
				return false;
		}
		visited.clear();//clean stack
		restart(); //load start position
		play_position = solution.size(); //set start play position
		return true;
	}

	/**
	*If map is solved do next step in solution.
	*
	*@return True if map is sloved and steps is done, false if map is not solved yet.
	*/
	bool play_solution_next() override {
		if (solution.empty()) return false; //puzzle is not solved yet
		if (play_position) {
			--play_position;
			move_sokoban(std::get<0>(solution[play_position]), std::get<1>(solution[play_position]));
		}
		return true;
	}

	/**
	*Gets type of cell.
	*
	*@param x The coordinate x of cell.
	*@param y The coordinate y of cell.
	*@return Byte number which represents field type according to the constants.
	*/
	element get_field_type(int x, int y) override{
		if (x == state.sokoban_x && y == state.sokoban_y)
			return MapUtils::make_sokoban_element(game_grid[y][x], SOKOBAN);
		return MapUtils::make_full_element(game_grid[y][x], state.barrel_grid[y][x]);
	}

	/**
	*Determines if in actual state of game is possible to move with all barrels.
	*
	*@return Bool value which determines if in actual state of game is possible to move with all barrels.
	*/
	bool is_dead_position() override;
	void right_move() override { move_sokoban(1, 0); }  /*!<Moves the sokoban to the right if it is possible. */
	void left_move() override { move_sokoban(-1, 0); } /*!<Moves the sokoban to the left if it is possible.*/
	void up_move() override { move_sokoban(0, -1); }  /*!<Moves the sokoban up if it is possible.*/
	void down_move() override { move_sokoban(0, 1); } /*!<Moves the sokoban down if it is possible.*/

	//only for two player map -> do nothing
	void right_move_second() override {} /*!<This method does nothing. There is no second sokoban*/
	void left_move_second() override {} /*!<This method does nothing. There is no second sokoban*/
	void up_move_second() override {} /*!<This method does nothing. There is no second sokoban*/
	void down_move_second() override {} /*!<This method does nothing. There is no second sokoban*/

private:
	std::vector<std::tuple<int, int>> solution; /*!<A sequence of steps defined by vector for solving puzzle*/
	GameState state; /*!<State of game including a position of sokoban and position of barrels.*/
	GameState start_state; /*!<Default state of game - for restart.*/
	int play_position; /*!< Variable determining which step from solution should be made while solution is playing.*/

	/**
	*General method for sokoban movement. The direction is defined by vector.
	*Tries to move sokoban in defined direction
	*
	*@param v_x The part x of vector which represents move direction of Sokoban. It may be -1,0 or 1.
	*@param v_y The part x of vector which represents move direction of Sokoban. It may be -1,0 or 1.
	*@return 0 if Sokoban was moved, -1 if Sokoban cannot be moved
	*/
	int move_sokoban(int v_x, int v_y);
	std::map<std::string, int> visited; /*!< The list of visited states (and number of steps which are visited in) for backtracking.*/

	/**
	*Tries to solve puzzle (backtracking) with at most MAX_STEPS. The solution is in vector solution.
	*
	*@param steps Number of steps from previous recursion step.
	*@return if solving succeed
	*/
	bool solve(int steps);
};
