#pragma once
#include "cinder/app/App.h"
#include "cinder/app/RendererGl.h"
#include "cinder/gl/gl.h"
#include "cinder/params/Params.h"
#include "SinglePlayerMap.h" //singleplayermap
#include "MyTools.h" //str_to_int, split
#include "cinder/Log.h"
#include "Timer.h" //Timer
#include "User.h" //User
#include "Map.h" //Map, TwoPlayerMap
#include "HighScore.h" //HighScore
#include <fstream> //ifstream, ofstream
#include <string> //string
#include <vector> //vector
#include <algorithm> //min
#include<memory> //unique pointer
#include<stdexcept> //exceptions
#include "MyExc.h" //corrupted_file_error
#include "Windows.h" //set working directory


/**
*Class representing game window. It is responsible to creating game environment and to interaction with keys.
*/
class SokobanApp : public ci::app::App {
public:

	/*
	*Creates game environment and menu. Loads highscore and information about the progress of users.
	*/
	void setup() override;

	/*
	*Method for drawing. Draws game grid and menu.
	*/
	void draw() override;

	/*
	*Method which is responsible to key interaction.
	*
	*@param event Represents key pressing.
	*/
	void keyDown(ci::app::KeyEvent event) override;
	
	/*
	*Constructor. Sets basic variables to default value.
	*/
	SokobanApp();

private:
	~SokobanApp(); /*!< Destructor. Writes highscore and the data of users to files.*/

	/** 
	*  @name Constans used in application.
	*  @{
	*/
	static const float FONT_SIZE; /*!< Size of font for drawing game information.*/
	static const float FIELD_SIZE_X; /*!< Width (in pixels) of the cell in game grid which represents a game element (ground, wall, barrel etc.)*/
	static const float FIELD_SIZE_Y; /*!< Height (in pixels) of the cell in game grid which represents a game element (ground, wall, barrel etc.)*/
	static const float MENU_X; /*!< Width (in pixels) of the menu.*/
	static const float MENU_Y; /*!< Height (in pixels) of the menu.*/
	static const float MENU_OFFSET; /*!< Distance between the menu and the game grid and between the menu and the border of window.*/
	static const float TEXT_BOX_HEIGHT; /*!< Height of all words which are drawed.*/
	static const float STATUS_WORD_DEFAULT_LENGTH; /*!< Default length of the word which describes a game status.*/
	static const char MAP_SEPARATOR; /*!< Character which must be at the end of map file.*/
	static const char USERS_SETINGS_SEPARATOR; /*!< Character which separates information for users in data file.*/ 
	static const char MAP_ID_SEPARATOR; /*!< Character which separates name of map pack and the number of map which is played. This two strings are made map id for highscore.*/
	static const int MAX_MAP_PACK_ID_LENGTH; /*!< Maximal length of the map pack id.*/
	static const char USER_SOURCE_FILE[]; /*!< Name of the source file for user data (proress).*/
	static const char HIGHSCORE_SOURCE_FILE[]; /*!< Name of the source file for highscore data.*/
	static const char MAP_HIHGSCORE_LABEL[]; /*!< Label which is drawed before map highscore.*/
	static const char USER_HIGHSCORE_LABEL[]; /*!< Label which is drawed before user highscore.*/
	static const char MAP_PACK_LOADED_LABEL[]; /*!< Label which is drawed before the current map pack name.*/
	static const char NUMBER_OF_MAP_LOADED_LABEL[]; /*!< Label which is drawed before the current map index within map pack.*/
	 
	static const std::vector<char> FORBIDDEN_MAP_ID_CHARS; /*!< Vectors of the forbidden symbols in map id.*/
	static const std::vector<char> FORBIDDEN_USER_NAME_CHARS; /*!< Vectors of the forbidden symbols in user name.*/
	static const std::vector<std::string> extensions; /*!< Vector of prefered extensions for map pack files.*/
	/** @} */ // end of group1


	/** 
	*  @name This is the group of the functions for buttons.
	*  @{
	*/
	void button_load(); /*!< Function for loading map pack (button load in menu).*/
	void button_restart(); /*!< Function for restarting current map (button restart in menu).*/
	void button_solve(); /*!< Function for the button solve (solving current map)*/
	void button_play_solution(); /*!< Funcion for the button which do next step in the solution of map.*/
	void button_login(); /*!< Funcion for the logging in the user (button login).*/
	void button_previous_map(); /*!< Function for the loading the previous map from the map pack (button previous map).*/
	void button_next_map(); /*!< Function for the loading the next map from the map pack (button next map).*/
	/** @} */ // end of group2

	/** 
	*  @name Tool functions. For settings app parameters, drawing map, etc.
	*  @{
	*/
	void set_new_map_params(); /*!< Sets parametrs (strings, size, etc.) for the current loaded map.*/
	void set_no_map_params(); /*!< Sets parametrs (strings, size, etc.) for the situation when no map is loaded.*/
	void set_high_score_strings(); /*!< Set highscore for the current loaded map.*/
	void draw_map(); /*!< Drawa the current loaded map (game grid).*/
	void initialize_users_progress(); /*!< Loads users progress from the data file.*/
	void initialize_highscore(); /*!< Loads highscore database from the data file.*/
	void initialize_textures(); /*!< Loads images representing game emelemts (ground, wall, etc.) as textures.*/
	void create_menu(); /*!< Creates menu (adds buttons, etc.).*/
	/** @} */ // end of group3

	/** 
	*  @name Variables defining current game status such as current loaded map, current user, sizes, etc.
	*  @{
	*/
	bool keys_disabled; /*!< Bool variable which determines if keys are disabled.*/
	bool new_map_set; /*!< Bool variable which determines if the new map was loaded.*/
	bool complete; /*!< Bool variable which determines if the map (puzzle) is solved by user.*/
	bool solved; /*!< Bool variable which determines if solver have solved the map.*/
	//size
	int s_x; /*!< Width of current map which is drawed (number of cells)*/
	int s_y; /*!< Width of current map which is drawed (number of cells)*/
	//for map and user
	std::string map_pack_id; /*<! Identifier of the current loaded map pack.*/
	int current_map_index; /*!< Integer variable representing which map from the list of maps is played.*/
	std::string m_user; /*!< String representing a content of text box for user logging in*/
	User * current_user; /*!< Pointer to user which is logged in. */
	/** @} */ // end of group4

	/** 
	*  @name Strings which are drawed in app including highscores, status line, map name.
	*  @{
	*/
	std::string status_line_word; /*!< String which contains the word describing the current game status.*/
	std::string map_best_word; /*!< String which contains best highscore for the current map.*/
	std::string user_best_word; /*!< String which contains the best highscore for the current user.*/
	std::string map_pack_name_word; /*!< String which contains name of the current loaded map pack.*/
	std::string user_name_word;  /*!< not used*/
	std::string game_mode_word; /*!< String which contains "single player" or "multi player" according to the game mode.*/
	/** @} */ // end of group5

	/** 
	*  @name Variables representing an engine of the game including list of maps, highscore database, timer, list of users
	*  @{
	*/
	std::vector<std::unique_ptr<Map>> list_of_maps; /*<! Vector of maps in the current loaded map pack.*/
	ci::params::InterfaceGlRef	mParams; /*!< Feature of Cinder library which represents a menu in this app.*/
	ci::gl::Texture2dRef textures[SinglePlayerMap::NUMBER_OF_FIELDS]; /*!< Array of textures which represents game elemets.*/
	MyTimer timer; /*!< Timer which measures the length of finding solution*/
	std::vector<User> available_users; /*!< Vector of users which have ever played the game.*/
	HighScore highScore; /*!< Highscore database.*/
	/** @} */
};