#include "Map.h" //Map, MapUtils


bool MapUtils::is_walkable_element(element el) {
	return el < Map::BARREL; //only ground or target is walkable
}

bool MapUtils::is_barrel_movable_element(element el, int barrel_info) {
	//check if it is possible to move barrel (el + barrel barrel_info <= 1 (TARGET)) - barrel grid may be 1 => *2
	return Map::TARGET >= el + (barrel_info << 1);
}

element MapUtils::make_full_element(element el, int barrel_info) {
	return el + (barrel_info > 0 ? Map::BARREL : 0);
}

element MapUtils::make_sokoban_element(element el, element sokoban_el) {
	return el + sokoban_el;
}

bool MapUtils::is_dead_position_triangle(element el_1, int b_1, element el_2, int b_2, element el_3, int b_3)
{
	if (b_1)
		b_1 = Map::BARREL;
	if (b_2)
		b_2 = Map::BARREL;
	if (b_3)
		b_3 = Map::BARREL;
	int p_1 = (el_1 >> 1) + b_1;  //3 for wall or 2 for barrel or 0 for nothing
	int p_2 = (el_2 & Map::BARREL) | b_2; //2 for barrel or wall 0 for nothing
	int p_3 = (el_3 >> 1) + b_3; //3 for wall or 2 for barrel or 0 for nothing
	if (p_1 + p_2 + p_3 > 5)
		return true; //dead position

	return false;
}

int MapUtils::number_of_new_barrel_on_target(element el_of_new_barrel, element el_of_old_barrel) {
	return  el_of_new_barrel - el_of_old_barrel;
}


Map::Map(std::ifstream & map_stream) : number_of_barrels(0), width(0), height(0) {
	int n_of_lines = 0;
	int n_of_columns = 0;
	bool first_line_read = false;
	game_grid.emplace_back();
	init_state.barrel_grid.emplace_back();
	char c;
	while (1) { // loop getting single characters
		if (!map_stream.get(c)) { 
			throw corrupted_file_error("map source file corrupted");
			return;
		}
		if (EOM == c) //end of map 
			break;
		if ('\n' == c) { //line readed
			if (first_line_read) {
				if (width != n_of_columns) //bad input
					throw corrupted_file_error("map source file corrupted");
			}
			else {
				width = n_of_columns;
				first_line_read = true;
			}
			++n_of_lines;
			game_grid.emplace_back();
			init_state.barrel_grid.emplace_back();
			n_of_columns = 0;
		}
		else {
			char f = c - '0';
			if (f < 0 || f >= NUMBER_OF_FIELDS) //bad input
				throw corrupted_file_error("map source file corrupted");
			if (SOKOBAN == f || SOKOBAN_TARGET == f) {
				init_state.sokoban_x = n_of_columns;
				init_state.sokoban_y = n_of_lines;
				game_grid[n_of_lines].push_back(f - SOKOBAN); //put 0 or 1 (ground or target)
				init_state.barrel_grid[n_of_lines].push_back(GROUND); //no barrel at this position
				init_state.hash.push_back(SOKOBAN);
			}
			else if (SOKOBAN_2 == f || SOKOBAN_2_TARGET == f) {
				init_state.sokoban2_x = n_of_columns;
				init_state.sokoban2_y = n_of_lines;
				game_grid[n_of_lines].push_back(f - SOKOBAN_2); //put 0 or 1 (ground or target)
				init_state.barrel_grid[n_of_lines].push_back(GROUND); //no barrel at this position
				init_state.hash.push_back(SOKOBAN_2); //or sokoban, if it is not used in get_field_type method 
			}
			else if (BARREL == f || BARREL_TARGET == f) {
				game_grid[n_of_lines].push_back(f - BARREL); //put 0 or 1 (ground or trget)
				init_state.barrel_grid[n_of_lines].push_back(++number_of_barrels); //barrel at this position
				init_state.barrels_on_target += f - BARREL; //BARREL - BARREL = 0, BARRELTARGET - BARREL = 1
				init_state.list_of_barrels.push_back(Barrel(n_of_columns, n_of_lines)); //add barrel with its coordinates
				init_state.hash.push_back(BARREL);
			}
			else {
				game_grid[n_of_lines].push_back(f);
				init_state.barrel_grid[n_of_lines].push_back(GROUND);
				init_state.hash.push_back(GROUND);
			}
			++n_of_columns;
		}
	}
	init_state.barrel_grid.pop_back(); //remove the last empty row
	height = n_of_lines; //set height
}


TwoPlayerMap::TwoPlayerMap(std::ifstream & map_stream) :Map(map_stream) {
	//check number of sokobans
	if (init_state.sokoban_x != -1 && init_state.sokoban2_x != -1) { //everything ok -> copy init state to ordinary game state
		state.sokoban_x = init_state.sokoban_x;
		state.sokoban_y = init_state.sokoban_y;
		state.sokoban2_x = init_state.sokoban2_x;
		state.sokoban2_y = init_state.sokoban2_y;
		state.barrels_on_target = init_state.barrels_on_target;
		state.barrel_grid = std::move(init_state.barrel_grid);
		//state.list_of_barrels = std::move(init_state.list_of_barrels);

		init_state.hash.clear(); // hash is not need
		init_state.list_of_barrels.clear(); //list of barrels is no needed
		//set default configuration
		start_state = state;
	}
	else {
		throw corrupted_file_error("to few sokobans");
	}
}


int TwoPlayerMap::move_sokoban(int v_x, int v_y, bool first_sokoban) {
	//positions of sokoban which will be moved
	int & m_sokoban_x = first_sokoban ? state.sokoban_x : state.sokoban2_x;
	int & m_sokoban_y = first_sokoban ? state.sokoban_y : state.sokoban2_y;
	//positions of the other
	int & o_sokoban_x = first_sokoban ? state.sokoban2_x : state.sokoban_x;
	int & o_sokoban_y = first_sokoban ? state.sokoban2_y : state.sokoban_y;
	//compute potential new position of sokoban
	int new_x = m_sokoban_x + v_x; //new position x of sokoban
	int new_y = m_sokoban_y + v_y; //new position y of sokoban
	if (!is_valid_position(new_x, new_y)) return false; //cannot move on this position, index overflow
	if (o_sokoban_x == new_x && o_sokoban_y == new_y) return -1;  //cannot move at this position, the another sokoban
	if (!MapUtils::is_walkable_element(game_grid[new_y][new_x])) return -1; //cannot move at this position, not walkable element
	if (state.barrel_grid[new_y][new_x]) { //there is a barrel on this position
		//compute position of potential new barrel position
		int new_barrel_x = new_x + v_x; //new position x of barrel
		int new_barrel_y = new_y + v_y; //new position y of barrel
		if (!is_valid_position(new_barrel_x, new_barrel_y)) return false; //cannot move barrel on this position, index overflow
		if (!MapUtils::is_barrel_movable_element(game_grid[new_barrel_y][new_barrel_x], state.barrel_grid[new_barrel_y][new_barrel_x])) return -1; //cannot move barrel at this position 
		if (o_sokoban_x == new_barrel_x && o_sokoban_y == new_barrel_y) return -1; //cannot move barrel at this position -> there is the another sokoban
		//move barrel
		std::swap(state.barrel_grid[new_barrel_y][new_barrel_x], state.barrel_grid[new_y][new_x]);
		//update the number of barrels on target 
		state.barrels_on_target += MapUtils::number_of_new_barrel_on_target(game_grid[new_barrel_y][new_barrel_x], game_grid[new_y][new_x]);
	}
	//now it is possible to move sokoban
	m_sokoban_x = new_x;
	m_sokoban_y = new_y;
	return 0; //success
}
