#include "SinglePlayerMap.h"

SinglePlayerMap::SinglePlayerMap(std::ifstream & map_stream) : Map(map_stream), play_position(0) {
	if (init_state.sokoban_x != -1 && init_state.sokoban2_x == -1) { //everything is ok -> copy init state to ordinary game state
		state.sokoban_x = init_state.sokoban_x;
		state.sokoban_y = init_state.sokoban_y;
		state.barrels_on_target = init_state.barrels_on_target;
		state.barrel_grid = std::move(init_state.barrel_grid);
		state.list_of_barrels = std::move(init_state.list_of_barrels);

		state.hash = std::move(init_state.hash); // hash is needed
		//set default configuration
		start_state = state;
	}
	else {
		throw corrupted_file_error("wrong number of sokobans");
	}
}

int SinglePlayerMap::move_sokoban(int v_x, int v_y) {
	int new_x = state.sokoban_x + v_x; //new position x of sokoban
	int new_y = state.sokoban_y + v_y; //new position y of sokoban
	if (!is_valid_position(new_x, new_y)) return -1; //cannot move on this position, index overflow
	if (!MapUtils::is_walkable_element(game_grid[new_y][new_x])) return -1; //cannot move on this position, not walkable element
	if (state.barrel_grid[new_y][new_x]) { //there is a barrel on this position
		//compute position of potential new barrel position
		int new_barrel_x = new_x + v_x; //new position x of barrel
		int new_barrel_y = new_y + v_y; //new position y of barrel
		if (!is_valid_position(new_barrel_x, new_barrel_y)) return false; //cannot move barrel on this position, index overflow
		if (!MapUtils::is_barrel_movable_element(game_grid[new_barrel_y][new_barrel_x], state.barrel_grid[new_barrel_y][new_barrel_x])) return -1; //cannot move barrel at this position 
		//move barrel
		int hlp = state.barrel_grid[new_y][new_x];
		state.barrel_grid[new_barrel_y][new_barrel_x] = hlp;
		state.barrel_grid[new_y][new_x] = 0;
		//update the number of barrels on target 
		state.barrels_on_target += MapUtils::number_of_new_barrel_on_target(game_grid[new_barrel_y][new_barrel_x], game_grid[new_y][new_x]);
		--hlp; //get index to list of barrels and update new position
		state.list_of_barrels[hlp].x = new_barrel_x;
		state.list_of_barrels[hlp].y = new_barrel_y;
		//update information on hash
		state.hash[width * new_y + new_x] = GROUND;
		state.hash[width * new_barrel_y + new_barrel_x] = BARREL;
	}
	state.hash[width * state.sokoban_y + state.sokoban_x] = GROUND; //delete old sokoban
	//now it is possible to move sokoban
	state.sokoban_x = new_x;
	state.sokoban_y = new_y;
	//update information of sokoban on hash
	state.hash[width * new_y + new_x] = SOKOBAN; //add new sokoban
	return 0; //success
}

bool SinglePlayerMap::is_dead_position() {

	for (const Barrel & b : state.list_of_barrels) {
		//barrel is on the target, such barrel may be imobile
		if (TARGET == game_grid[b.y][b.x]) continue;
		//for corners check dead position
		for (int i = -1; i <= 1; i += 2) {
			for (int j = -1; j <= 1; j += 2) {
				element game_el_1;
				element game_el_2;
				element game_el_3;
				int barrel_1;
				int barrel_2;
				int barrel_3;
				if (is_valid_position(b.x + j, b.y)) {
					game_el_1 = game_grid[b.y][b.x + j];
					barrel_1 = state.barrel_grid[b.y][b.x + j];
				}
				else { //overflow is the same situaion as wall
					game_el_1 = WALL;
					barrel_1 = 0;
				}
				if (is_valid_position(b.x + j, b.y + i)) {
					game_el_2 = game_grid[b.y + i][b.x + j];
					barrel_2 = state.barrel_grid[b.y + i][b.x + j];
				}
				else { //overflow is the same situaion as wall
					game_el_2 = WALL;
					barrel_2 = 0;
				}
				if (is_valid_position(b.x, b.y + i)) {
					game_el_3 = game_grid[b.y + i][b.x ];
					barrel_3 = state.barrel_grid[b.y + i][b.x];
				}
				else { //overflow is the same situaion as wall
					game_el_3 = WALL;
					barrel_3 = 0;
				}				
				if (MapUtils::is_dead_position_triangle(game_el_1, barrel_1, game_el_2, barrel_2, game_el_3, barrel_3))
					return true; //dead position
			}
		}
	}
	return false;
}

bool SinglePlayerMap::solve(int steps) {
	if (complete()) //puzzle is complete
		return true;
	if (steps > MAX_STEPS) return false; //to much steps	
	if (is_dead_position())
		return false;
	if (visited.find(state.hash) != visited.end() && visited[state.hash] < steps) { //this vertex is visited in less steps
		return false;
	}
	visited[state.hash] = steps;
	GameState help = state; //save original position
	for (int i = -1; i <= 1; i += 2) {
		if (!move_sokoban(i, 0)) {
			if (solve(steps + 1)) {
				state = help;
				solution.push_back(std::tuple<int, int>(i, 0)); //push direction to the  vector
				return true;
			}
			state = help;
		}
		if (!move_sokoban(0, i)) {
			if (solve(steps + 1)) {
				state = help;
				solution.push_back(std::tuple<int, int>(0, i)); //push direction to the  vector
				return true;
			}
			state = help;
		}

	}
	return false;
}