﻿#include "SokobanApp.h"

//initialize constants
const float SokobanApp::FONT_SIZE = 20;
const float SokobanApp::FIELD_SIZE_X = 40;
const float SokobanApp::FIELD_SIZE_Y = 39;
const float SokobanApp::MENU_X = 150;
const float SokobanApp::MENU_Y = 300;
const float SokobanApp::MENU_OFFSET = 25;
const float SokobanApp::TEXT_BOX_HEIGHT = 80;
const float SokobanApp::STATUS_WORD_DEFAULT_LENGTH = 300;
const char SokobanApp::MAP_SEPARATOR = 'x';
const char SokobanApp::USERS_SETINGS_SEPARATOR = ';';
const char SokobanApp::MAP_ID_SEPARATOR = '+';
const int SokobanApp::MAX_MAP_PACK_ID_LENGTH = 20;
const char SokobanApp::USER_SOURCE_FILE[] = "users.dat";
const char SokobanApp::HIGHSCORE_SOURCE_FILE[] = "highscore.dat";
const char SokobanApp::MAP_HIHGSCORE_LABEL[] = "map highscore: ";
const char SokobanApp::USER_HIGHSCORE_LABEL[] = "user highscore: ";
const char SokobanApp::MAP_PACK_LOADED_LABEL[] = "Loaded map pack: ";
const char SokobanApp::NUMBER_OF_MAP_LOADED_LABEL[] = "Map: ";

const std::vector<char> SokobanApp::FORBIDDEN_MAP_ID_CHARS = { '+', ';' };
const std::vector<char> SokobanApp::FORBIDDEN_USER_NAME_CHARS = { ';' };
const std::vector<std::string> SokobanApp::extensions = { "txt" };

SokobanApp::SokobanApp() : keys_disabled(true), new_map_set(false), complete(false), solved(false), s_x(0), s_y(0), map_pack_id(""),   current_map_index(-1), m_user(""),
							current_user(nullptr), status_line_word(""), map_best_word(""), user_best_word(""), map_pack_name_word(""), user_name_word(""), game_mode_word("") {}


void SokobanApp::button_load()
{
	//no user is logged -> return
	if (!current_user) {
		status_line_word = "no user logged in";
		return;
	}
	char current_working_dir[MAX_PATH]; //because getOpenFilePath changes current directory
	GetCurrentDirectoryA(MAX_PATH, current_working_dir);
	try {
		ci::fs::path path = getOpenFilePath("", extensions);
		if (!path.empty()) {
			std::ifstream map_pack_file; //input stream for map pack source file
			map_pack_file.open(path.c_str()); //open map pack source file
			std::string mode;
			//set map pack id and mode (two player or single player)
			if (!getline(map_pack_file, map_pack_id) || "" == map_pack_id ||
				!getline(map_pack_file, mode)) { 
				status_line_word = "map pack source file corrupted\nsome information missing";
				return;
			}
			//check if map_pack_id does not contain forbidden characters
			for (const char & fc : FORBIDDEN_MAP_ID_CHARS) {
				if (map_pack_id.find(fc) != std::string::npos) {
					status_line_word = "map pack id name cannot\ncontains ";
					status_line_word += fc;
					return;
				}
			}
			bool single_player; //determines if mode is single player
			if ("SP" == mode) single_player = true;
			else if ("MP" == mode) single_player = false;
			else {
				status_line_word = "map pack source file corrupted\nsome information missing";
				return;
			}
			std::ifstream map_file; //stream which reads from map source files
			std::string map_file_name; //name of source file
			list_of_maps.clear(); //clear previous list of maps
			timer.stop_clock(); // stop clock
			while (getline(map_pack_file, map_file_name)) { 
				map_file.open(map_file_name);
				if (!map_file.is_open()) {
					throw corrupted_file_error("map source file does not exist");
				}
				if (single_player) {//add map according to the mode
					list_of_maps.push_back(std::make_unique<SinglePlayerMap>(map_file));
					game_mode_word = " (single player)";
				}
				else {
					list_of_maps.push_back(std::make_unique<TwoPlayerMap>(map_file));
					game_mode_word = " (multi player)";
				}
					
				map_file.close();
			}
			if (list_of_maps.empty()) //no map loaded
				throw corrupted_file_error("map pack source file corrupted");
			current_map_index = current_user->get_progress(map_pack_id);
			set_new_map_params();
			set_high_score_strings();
			timer.start_clock();
			keys_disabled = false;
		}
	}
	catch (const corrupted_file_error & e) {
		if (list_of_maps.empty()) { //no map loaded
			set_no_map_params();
			keys_disabled = true;
			status_line_word = e.what(); //type error to status line
			status_line_word += "\ncannot load any map";
			return;
		}
		else { //some of maps loaded
			//minimum of progress and number of loaded map - progress is always bigger then one -> it is ok to cast to unsigned int
			current_map_index = std::min(static_cast<std::size_t>(current_user->get_progress(map_pack_id)), list_of_maps.size() - 1);
			set_new_map_params(); 
			set_high_score_strings();
			timer.start_clock();
			keys_disabled = false;
			status_line_word = e.what(); //type error to status line
			status_line_word += "\ncannot load some of maps";
		}
	}
	SetCurrentDirectoryA(current_working_dir);
}

void SokobanApp::button_restart()
{
	if (-1 == current_map_index) {//no map loaded
		status_line_word = "no map loaded";
		return;
	}
	list_of_maps[current_map_index]->restart();
	keys_disabled = false;
	solved = false;
	status_line_word = "";
	timer.start_clock();
}

void SokobanApp::button_solve() {
	if (-1 == current_map_index) { //no map loaded
		status_line_word = "no map loaded";
		return;
	}

	if (list_of_maps[current_map_index]->solve()) {
		status_line_word = "solved";
		keys_disabled = true; //disable key, it is not allowed to move during playing solution
		solved = true; //playing solution is possible
		timer.stop_clock(); //clock are unnecessary to play solution
	}
	else {
		status_line_word = "cannot solve map";
	}
}

void SokobanApp::button_play_solution() {
	if (solved) //it is possible to play solution
		list_of_maps[current_map_index]->play_solution_next();
	else
		status_line_word = -1 < current_map_index ? "solve puzzle first" : "no map loaded";
}

void SokobanApp::button_login() {
	//empty text field
	if ("" == m_user) {
		return;
	}

	//check if username does not contain forbidden characters
	for (const char & fc : FORBIDDEN_USER_NAME_CHARS) {
		if (m_user.find(fc) != std::string::npos) {
			status_line_word = "user name cannot\ncontains ";
			status_line_word += fc;
			return;
		}
	}

	User logged_user(m_user); //create user
	auto it = std::find(available_users.begin(), available_users.end(), logged_user); //try to find user in list
	//new user add him to list of available users
	if (it == available_users.end()) {
		available_users.push_back(logged_user);
		current_user = &available_users.back();
	}
	else {
		current_user = &*it;
	}
	if (0 != list_of_maps.size()) { //if map is loaded
		current_map_index = current_user->get_progress(map_pack_id);
		set_new_map_params();
		set_high_score_strings();
		timer.start_clock();
		keys_disabled = false;
	}
	status_line_word = m_user + " logged successfully";
}

void SokobanApp::setup() {
	//create menu
	create_menu(); 
	//try to load textures
	try {
		initialize_textures();
	}
	catch (const std::exception & e) {
		status_line_word += e.what();
		status_line_word += '\n';
	}
	//set size
	ci::app::setWindowSize(static_cast<int>(MENU_X + std::max(STATUS_WORD_DEFAULT_LENGTH, MENU_OFFSET + s_x * FIELD_SIZE_X)), static_cast<int>(std::max( MENU_Y + MENU_OFFSET, s_y * FIELD_SIZE_Y)));
	initialize_users_progress();
	initialize_highscore();
}

void SokobanApp::keyDown(ci::app::KeyEvent event) {
	if (keys_disabled)//keys are disabled -> return
		return;
	//determine which key was pressed
	switch (event.getCode()) {
	case ci::app::KeyEvent::KEY_RIGHT:
		list_of_maps[current_map_index]->right_move();
		break;
	case ci::app::KeyEvent::KEY_LEFT:
		list_of_maps[current_map_index]->left_move();
		break;
	case ci::app::KeyEvent::KEY_UP:
		list_of_maps[current_map_index]->up_move();
		break;
	case ci::app::KeyEvent::KEY_DOWN:
		list_of_maps[current_map_index]->down_move();
		break;
	case ci::app::KeyEvent::KEY_d:
		list_of_maps[current_map_index]->right_move_second();
		break;
	case ci::app::KeyEvent::KEY_a:
		list_of_maps[current_map_index]->left_move_second();
		break;
	case ci::app::KeyEvent::KEY_w:
		list_of_maps[current_map_index]->up_move_second();
		break;
	case ci::app::KeyEvent::KEY_s:
		list_of_maps[current_map_index]->down_move_second();
		break;
	default:
		break;
	}
	//check if map is complete
	if (list_of_maps[current_map_index]->complete()) {
		keys_disabled = true; //disable key
		status_line_word = "Level completed!";
		uint64_t sec = timer.stop_clock();	
		if (static_cast<std::size_t>(current_map_index)  < list_of_maps.size() - 1) //update progress for user
			current_user->set_progress(map_pack_id, current_map_index + 1);
		std::string map_id = map_pack_id + MAP_ID_SEPARATOR + std::to_string(current_map_index); //"compute map id"
		HighScore::result res =  highScore.set_time(current_user->get_name(), map_id, sec); //update highscore
		if (HighScore::result::best_user_time == res)
			status_line_word += "\nuser best";
		else if (HighScore::result::best_map_time == res)
			status_line_word += "\nmap best";
		set_high_score_strings();
	}
	if (list_of_maps[current_map_index]->is_dead_position())
		status_line_word = "dead position";
}


/*void SokobanApp::update()
{
}*/

void SokobanApp::draw()
{
	ci::gl::setMatricesWindow(getWindowSize());
	ci::gl::clear(ci::Color(0, 0, 0));
	//draw time
	ci::gl::translate(MENU_OFFSET, MENU_Y + MENU_OFFSET);
	ci::gl::drawString(timer.get_time(), ci::ivec2(0, 0), ci::Color::white(), ci::Font("Helvetica Neue", 1.5*FONT_SIZE));
	//draw header
	ci::gl::translate(MENU_X , -1 * MENU_Y);
	ci::gl::drawString(MAP_PACK_LOADED_LABEL + map_pack_id + game_mode_word, ci::ivec2(0, 0), ci::Color::white(), ci::Font("Helvetica Neue", FONT_SIZE ));
	ci::gl::translate(0, FONT_SIZE);
	ci::gl::drawString(NUMBER_OF_MAP_LOADED_LABEL + (-1 == current_map_index ? "" :
		std::to_string(current_map_index + 1) + '/' + std::to_string(list_of_maps.size())),
		ci::ivec2(0, 0), ci::Color::white(), ci::Font("Helvetica Neue", FONT_SIZE));
	ci::gl::translate(0, FONT_SIZE);
	//draw map
	draw_map();
	//draw menu
	mParams->draw();
	//draw status line
	ci::gl::drawString(MAP_HIHGSCORE_LABEL + map_best_word, ci::ivec2(0, 0), ci::Color::white(), ci::Font("Helvetica Neue", FONT_SIZE));
	ci::gl::translate(0, FONT_SIZE);
	ci::gl::drawString(USER_HIGHSCORE_LABEL + user_best_word, ci::ivec2(0, 0), ci::Color::white(), ci::Font("Helvetica Neue", FONT_SIZE));
	ci::gl::translate(0, FONT_SIZE);
	ci::gl::drawString(status_line_word, ci::ivec2(0, 0), ci::Color::white(), ci::Font("Helvetica Neue", FONT_SIZE * 1.3));
}

void SokobanApp::draw_map() {
	if (new_map_set) {
		setWindowSize(static_cast<int>(MENU_X + MENU_OFFSET + s_x * FIELD_SIZE_X), static_cast<int>(std::max(MENU_Y + MENU_OFFSET + FONT_SIZE, s_y * FIELD_SIZE_Y + TEXT_BOX_HEIGHT*2)));
		new_map_set = false;
	}

	for (int j = 0; j < s_y; j++)
	{
		for (int i = 0; i < s_x; i++)
		{
			ci::gl::draw(textures[list_of_maps[current_map_index]->get_field_type(i, j)]);
			ci::gl::translate(FIELD_SIZE_X, 0);
		}
		ci::gl::translate(FIELD_SIZE_X * -1 * s_x, FIELD_SIZE_Y);
	}

}

void SokobanApp::initialize_users_progress() {
	std::ifstream users_settings_file;
	users_settings_file.open(USER_SOURCE_FILE);
	if (!users_settings_file.is_open()) //some problem, maybe file does not exist (no progress yet)
		return;
	std::string line; //line of input file
	std::vector<std::string> splitted_line; //splitted line of input file
	while (std::getline(users_settings_file, line)) { //read lines
		MyTools::split(line, USERS_SETINGS_SEPARATOR, splitted_line);
		if (splitted_line.size() < 3) { //to few data
			status_line_word += "users data file corrupted\n";
			continue;
		}
		available_users.emplace_back(splitted_line[0]);
		int prog;
		for (std::size_t i = 1; i + 1 < splitted_line.size(); i = i + 2) {
			try {
				prog = MyTools::str_to_int(splitted_line[i+1]);
				available_users.back().set_progress(splitted_line[i], prog);
			}
			catch (const std::invalid_argument &) { //cannot split line
				status_line_word += "users data file corrupted\n";
				continue;
			}
		}
		splitted_line.clear();
	}
}

void SokobanApp::initialize_highscore() {
	std::ifstream highscore_file;
	highscore_file.open(HIGHSCORE_SOURCE_FILE);
	if (!highscore_file.is_open()) //some problem, maybe file does not exist (no highscore yet)
		return;
	std::string line; //line of input file
	std::vector<std::string> splitted_line; //splitted line of input file
	while (std::getline(highscore_file, line)) { //read lines
		MyTools::split(line, USERS_SETINGS_SEPARATOR, splitted_line);
		if (splitted_line.size() < 3) { //to few data
			status_line_word += "highscore data file corrupted\n";
			continue;
		}
		int time_sec;
		for (std::size_t i = 1; i + 1 < splitted_line.size(); i += 2) {
			try {
				time_sec = MyTools::str_to_int(splitted_line[i + 1]);
				highScore.set_time(splitted_line[0],splitted_line[i], time_sec);
			}
			catch (const std::invalid_argument &) { //cannot split line
				status_line_word += "highscore data file corrupted\n";
				continue;
			}
		}
		splitted_line.clear();
	}
}

void SokobanApp::initialize_textures() {
	auto img = loadImage(loadAsset("grass.png"));
	textures[SinglePlayerMap::GRASS] = ci::gl::Texture2d::create(img);
	//load and set wall texture
	img = loadImage(loadAsset("wall.png"));
	textures[SinglePlayerMap::WALL] = ci::gl::Texture2d::create(img);
	//load and set sokoban texture
	img = loadImage(loadAsset("sokoban.png"));
	textures[SinglePlayerMap::SOKOBAN] = ci::gl::Texture2d::create(img);
	//load and set sokoban on target texture
	img = loadImage(loadAsset("sokoban_target.png"));
	textures[SinglePlayerMap::SOKOBAN_TARGET] = ci::gl::Texture2d::create(img);
	//load and set ground texture
	img = loadImage(loadAsset("ground.png"));
	textures[SinglePlayerMap::GROUND] = ci::gl::Texture2d::create(img);
	//load and set target texture
	img = loadImage(loadAsset("target.png"));
	textures[SinglePlayerMap::TARGET] = ci::gl::Texture2d::create(img);
	//load and set barrel texture
	img = loadImage(loadAsset("barrel.png"));
	textures[SinglePlayerMap::BARREL] = ci::gl::Texture2d::create(img);
	//load and set barrel on target texture
	img = loadImage(loadAsset("barrel_target.png"));
	textures[SinglePlayerMap::BARREL_TARGET] = ci::gl::Texture2d::create(img);
	//load and set sokoban 2 texture
	img = loadImage(loadAsset("sokoban2.png"));
	textures[SinglePlayerMap::SOKOBAN_2] = ci::gl::Texture2d::create(img);
	//load and set sokoban 2 on target texture
	img = loadImage(loadAsset("sokoban2_target.png"));
	textures[SinglePlayerMap::SOKOBAN_2_TARGET] = ci::gl::Texture2d::create(img);
}

void SokobanApp::create_menu() {
	//add "menu"
	mParams = ci::params::InterfaceGl::create(getWindow(), "MENU", ci::app::toPixels(ci::ivec2(MENU_X, MENU_Y)));
	//add start button
	mParams->addButton("Load map", std::bind(&SokobanApp::button_load, this));
	//add reset button
	mParams->addButton("Reset level", std::bind(&SokobanApp::button_restart, this));
	//add separator
	mParams->addSeparator();
	//add solve button
	mParams->addButton("Solve", std::bind(&SokobanApp::button_solve, this));
	//add next button
	mParams->addButton("next", std::bind(&SokobanApp::button_play_solution, this));
	//add login issues
	mParams->addSeparator();
	mParams->addParam("User", &m_user);
	mParams->addButton("login", std::bind(&SokobanApp::button_login, this));
	//mParams->removeParam("login");
	mParams->addSeparator();
	mParams->addButton("next map", std::bind(&SokobanApp::button_next_map, this));
	mParams->addButton("previous map", std::bind(&SokobanApp::button_previous_map, this));
}

void SokobanApp::button_next_map() {
	if (!current_user || -1 == current_map_index) {
		status_line_word = "no map loaded";
		return;
	}

	if (static_cast<std::size_t>(current_map_index) >= list_of_maps.size() - 1) { //max level reached, 
		status_line_word = "This is the last map";
		return;
	}
	else if (current_map_index == current_user->get_progress(map_pack_id)) {
		status_line_word = "solve map first";
		return;
	}

	list_of_maps[current_map_index]->restart(); //exit (restart) current map
	++current_map_index; //move to the next map
	set_new_map_params(); //do routine for new map set
	set_high_score_strings(); //set highscores
	timer.start_clock(); //restart clock
	keys_disabled = false; //allow movement
}

void SokobanApp::button_previous_map() {
	if (current_map_index <= 0) return; //no previous map
	list_of_maps[current_map_index]->restart(); //exit (restart) current map
	--current_map_index; //move to the previous map
	set_new_map_params(); //do routine for new map set
	set_high_score_strings(); //set highscores
	timer.start_clock(); //restart clock
	keys_disabled = false; //allow movement
}

void SokobanApp::set_new_map_params() {
	list_of_maps[current_map_index]->restart();
	s_x = list_of_maps[current_map_index]->get_width();
	s_y = list_of_maps[current_map_index]->get_height();
	new_map_set = true;
	complete = false;
	solved = false;
	status_line_word = "";
	//timer.start_clock();
}

void SokobanApp::set_no_map_params() {
	s_x = 0;
	s_y = 0;
	current_map_index = -1;
}

SokobanApp::~SokobanApp() {
	//write progress to file
	std::ofstream users_settings_file;
	users_settings_file.open(USER_SOURCE_FILE);
	for (auto const& x : available_users) {
		users_settings_file << x << std::endl;
	}
	//write highscore to file
	std::ofstream highscore_data_file;
	highscore_data_file.open(HIGHSCORE_SOURCE_FILE);
	highScore.to_stream(highscore_data_file);
}

void SokobanApp::set_high_score_strings() {
	std::string map_id = map_pack_id + MAP_ID_SEPARATOR + std::to_string(current_map_index); //"compute map id"
	if ("" != map_pack_id) //check if map is set
		map_best_word = highScore.get_map_best_time(map_id); //get highscore string
	else
		map_best_word = ""; 
	if (current_user) //if user is logged
		user_best_word = highScore.get_user_best_time(current_user->get_name(), map_id);
	else
		user_best_word = "";

}

//makro which is expanded to the main function
CINDER_APP(SokobanApp, ci::app::RendererGl )
